@@ -59,8 +59,12 @@ FTS_DBUS_INTERFACE = "org.gnome.zeitgeist.Index"
 FILTER_PREFIX_EVENT_INTERPRETATION = "ZGEI"
 FILTER_PREFIX_EVENT_MANIFESTATION = "ZGEM"
 FILTER_PREFIX_ACTOR = "ZGA"
+FILTER_PREFIX_SUBJECT_URI = "ZGSU"
 FILTER_PREFIX_SUBJECT_INTERPRETATION = "ZGSI"
 FILTER_PREFIX_SUBJECT_MANIFESTATION = "ZGSM"
+FILTER_PREFIX_SUBJECT_ORIGIN = "ZGSO"
+FILTER_PREFIX_SUBJECT_MIMETYPE = "ZGST"
+FILTER_PREFIX_SUBJECT_STORAGE = "ZGSS"
 FILTER_PREFIX_XDG_CATEGORY = "AC"
 
 VALUE_EVENT_ID = 0
@@ -129,6 +133,36 @@ class SearchEngineExtension (Extension, dbus.service.Object):
 				event._make_dbus_sendable()
 		return [NULL_EVENT if event is None else event for event in events]
 
+def mangle_uri (uri):
+	"""
+	Converts a URI into an index- and query friendly string. The problem
+	is that Xapian doesn't handle CAPITAL letters or most non-alphanumeric
+	symbols in a boolean term when it does prefix matching. The mangled
+	URIs returned from this function are suitable for boolean prefix searches.
+	
+	IMPORTANT: This is a 1-way function! You can not convert back.
+	"""
+	result = ""
+	for c in uri.lower():
+		if c in (": /"):
+			result += "_"
+		else:
+			result += c
+	return result
+
+def expand_type (type_prefix, uri):
+	"""
+	Return a string with a Xapian query matching all child types of 'uri'
+	inside the Xapian prefix 'type_prefix'.
+	"""
+	is_negation = uri.startswith(NEGATION_OPERATOR)
+	uri = uri[1:] if is_negation else uri
+	children = Symbol.find_child_uris_extended(uri)
+	children = [ "%s:%s" % (type_prefix, child) for child in children ]
+
+	result = " OR ".join(children)
+	return result if not is_negation else "NOT (%s)" % result
+
 class Indexer:
 	"""
 	Abstraction of the FT indexer and search engine
@@ -155,8 +189,12 @@ class Indexer:
 		self._query_parser.add_boolean_prefix("zgei", FILTER_PREFIX_EVENT_INTERPRETATION)
 		self._query_parser.add_boolean_prefix("zgem", FILTER_PREFIX_EVENT_MANIFESTATION)
 		self._query_parser.add_boolean_prefix("zga", FILTER_PREFIX_ACTOR)
+		self._query_parser.add_prefix("zgsu", FILTER_PREFIX_SUBJECT_URI)
 		self._query_parser.add_boolean_prefix("zgsi", FILTER_PREFIX_SUBJECT_INTERPRETATION)
 		self._query_parser.add_boolean_prefix("zgsm", FILTER_PREFIX_SUBJECT_MANIFESTATION)
+		self._query_parser.add_prefix("zgso", FILTER_PREFIX_SUBJECT_ORIGIN)
+		self._query_parser.add_boolean_prefix("zgst", FILTER_PREFIX_SUBJECT_MIMETYPE)
+		self._query_parser.add_boolean_prefix("zgss", FILTER_PREFIX_SUBJECT_STORAGE)
 		self._query_parser.add_prefix("category", FILTER_PREFIX_XDG_CATEGORY)
 		self._query_parser.add_valuerangeprocessor(
 		      xapian.NumberValueRangeProcessor(VALUE_EVENT_ID, "id", True))
@@ -176,19 +214,35 @@ class Indexer:
 		self._worker.daemon = True
 		self._worker.start()
 		
+		self._check_index ()
+		
+	def _check_index (self):
 		if self._index.get_doccount() == 0:
+			# If the index is empty we trigger a rebuild
 			# We must delay reindexing until after the engine is done setting up
 			log.info("Empty index detected. Doing full rebuild")
 			gobject.idle_add (self._reindex)
-		
-		# List term freqs
-		#for term in self._index.allterms():
-		#	print term.term, term.termfreq
+		else:
+			# If the index doesn't use the zgsu prefix, it must be old-style,
+			# and we must rebuild it
+			query = self._query_parser.parse_query ("zgsu:file*",
+		                                            self.QUERY_PARSER_FLAGS)
+			self._enquire.set_query (query)
+			hits = self._enquire.get_mset (0, 1)
+			hit_count = hits.get_matches_estimated()
+			if hit_count == 0:
+				log.info ("Old index format detected. Rebuilding index.")
+				gobject.idle_add (self._reindex)
 	
 	def _reindex (self):
 		"""
 		Index everything in the ZG log
 		"""
+		self._index.close ()
+		self._index = xapian.WritableDatabase(INDEX_FILE, xapian.DB_CREATE_OR_OVERWRITE)
+		self._query_parser.set_database (self._index)
+		self._enquire = xapian.Enquire(self._index)
+		
 		all_events = self._engine.find_events(TimeRange.always(),
 		                                      [], StorageState.Any,
 		                                      sys.maxint,
@@ -426,7 +480,7 @@ class Indexer:
 			
 				doc = self._tokenizer.get_document()
 				for cat in desktop.getCategories():
-					doc.add_term(FILTER_PREFIX_XDG_CATEGORY+cat.lower())
+					doc.add_boolean_term(FILTER_PREFIX_XDG_CATEGORY+cat.lower())
 		else:
 			log.debug("Unable to look up app info for %s" % actor)
 		
@@ -505,13 +559,26 @@ class Indexer:
 	def _add_doc_filters (self, event, doc):
 		"""Adds the filtering rules to the doc. Filtering rules will
 		   not affect the relevancy ranking of the event/doc"""
-		doc.add_term (FILTER_PREFIX_EVENT_INTERPRETATION+event.interpretation)
-		doc.add_term (FILTER_PREFIX_EVENT_MANIFESTATION+event.manifestation)
-		doc.add_term (FILTER_PREFIX_ACTOR+event.actor)
+		if event.interpretation:
+			doc.add_boolean_term (FILTER_PREFIX_EVENT_INTERPRETATION+event.interpretation)
+		if event.manifestation:
+			doc.add_boolean_term (FILTER_PREFIX_EVENT_MANIFESTATION+event.manifestation)
+		if event.actor:
+			doc.add_boolean_term (FILTER_PREFIX_ACTOR+mangle_uri(event.actor))
 		
 		for su in event.subjects:
-			doc.add_term (FILTER_PREFIX_SUBJECT_INTERPRETATION+su.interpretation)
-			doc.add_term (FILTER_PREFIX_SUBJECT_MANIFESTATION+su.manifestation)
+			if su.uri:
+				doc.add_boolean_term (FILTER_PREFIX_SUBJECT_URI+mangle_uri(su.uri))
+			if su.interpretation:
+				doc.add_boolean_term (FILTER_PREFIX_SUBJECT_INTERPRETATION+su.interpretation)
+			if su.manifestation:
+				doc.add_boolean_term (FILTER_PREFIX_SUBJECT_MANIFESTATION+su.manifestation)
+			if su.origin:
+				doc.add_boolean_term (FILTER_PREFIX_SUBJECT_ORIGIN+mangle_uri(su.origin))
+			if su.mimetype:
+				doc.add_boolean_term (FILTER_PREFIX_SUBJECT_MIMETYPE+su.mimetype)
+			if su.storage:
+				doc.add_boolean_term (FILTER_PREFIX_SUBJECT_STORAGE+su.storage)
 	
 	def _index_event_real (self, event):
 		if not isinstance (event, Event):
@@ -535,29 +602,22 @@ class Indexer:
 				log.debug("Indexing '%s'" % subject.uri)
 				self._index_uri (subject.uri)
 				self._index_text (subject.text)
+				
+				# If the subject URI is an actor, we index the .desktop also
 				if subject.uri.startswith ("application://"):
-				  self._index_actor (subject.uri)
-				# FIXME: index origin into an origin: prefix
-				# FIXME: index uri into a uri: prefix
+					self._index_actor (subject.uri)
+				
+				# File contents indexing disabled for now...
 				#self._index_contents (subject.uri)
+				
 				# FIXME: Possibly index payloads when we have apriori knowledge
 			
 			self._add_doc_filters (event, doc)	
-				
 			self._index.add_document (doc)
+		
 		except Exception, e:
 			log.error("Error indexing event: %s" % e)
 
-	def _expand_type (self, type_prefix, uri):
-		is_negation = uri.startswith(NEGATION_OPERATOR)
-		uri = uri[1:] if is_negation else uri
-		children = Symbol.find_child_uris_extended(uri)
-		children = [ "%s:%s" % (type_prefix, child) for child in children ]
-
-		result = " OR ".join(children)
-
-		return result if not is_negation else "NOT (%s)" % result
-
 	def _compile_event_filter_query (self, events):
 		"""Takes a list of event templates and compiles a filter query
 		   based on their, interpretations, manifestations, and actor,
@@ -575,15 +635,23 @@ class Indexer:
 			
 			tmpl = []
 			if event.interpretation :
-				tmpl.append(self._expand_type("zgei", event.interpretation))
+				tmpl.append(expand_type("zgei", event.interpretation))
 			if event.manifestation :
-				tmpl.append(self._expand_type("zgem", event.manifestation))
-			if event.actor : tmpl.append("zga:"+event.actor)
+				tmpl.append(expand_type("zgem", event.manifestation))
+			if event.actor : tmpl.append("zga:%s" % mangle_uri(event.actor))
 			for su in event.subjects:
+				if su.uri :
+					tmpl.append("zgsu:%s" % mangle_uri(su.uri))
 				if su.interpretation :
-					tmpl.append(self._expand_type("zgsi", su.interpretation))
+					tmpl.append(expand_type("zgsi", su.interpretation))
 				if su.manifestation :
-					tmpl.append(self._expand_type("zgsm", su.manifestation))
+					tmpl.append(expand_type("zgsm", su.manifestation))
+				if su.origin :
+					tmpl.append("zgso:%s" % mangle_uri(su.origin))
+				if su.mimetype :
+					tmpl.append("zgst:%s" % su.mimetype)
+				if su.storage :
+					tmpl.append("zgss:%s" % su.storage)
 			
 			tmpl = "(" + ") AND (".join(tmpl) + ")"
 			query.append(tmpl)
